; stage1.asm
[BITS 16]
[ORG 0x7C00]
STAGE2_SEGMENT equ 0x0800   ; Load stage 2 at 0x8000
STAGE2_SECTORS equ 2        ; Size of stage 2 in sectors

start:
    ; Disable interrupts during setup
    cli
    ; Set up segment registers
    xor ax, ax
    mov ds, ax
    mov es, ax
    mov ss, ax
    mov sp, 0x7C00
    mov [boot_drive], dl    ; Store boot drive number
    
    ; Store drive parameters for USB detection
    push dx
    mov ah, 0x41            ; Check ext disk support
    mov bx, 0x55AA
    mov dl, [boot_drive]
    int 0x13
    jc .legacy_load         ; If not supported, use legacy loading
    cmp bx, 0xAA55
    jne .legacy_load
    test cl, 1              ; Test if ext disk access functions supported
    jz .legacy_load
    mov byte [ext_support], 1
    
.legacy_load:
    pop dx
    sti

    ; Load stage 2 using appropriate method
    cmp byte [ext_support], 1
    je .ext_load
    
    ; Legacy load method
    mov ax, STAGE2_SEGMENT
    mov es, ax
    xor bx, bx             ; ES:BX = 0x8000:0000
    mov ah, 0x02           ; Read sectors
    mov al, STAGE2_SECTORS
    mov ch, 0              ; Cylinder 0
    mov cl, 2              ; Start from sector 2
    mov dh, 0              ; Head 0
    mov dl, [boot_drive]
    int 0x13
    jc disk_error
    jmp .load_complete

.ext_load:
    ; Extended load method for USB drives
    push si
    push ds
    
    ; Set up disk address packet
    mov ax, 0
    mov ds, ax
    mov si, dap
    mov ah, 0x42
    mov dl, [boot_drive]
    int 0x13
    
    pop ds
    pop si
    jc disk_error

.load_complete:
    ; Pass boot info to stage 2
    mov dl, [boot_drive]
    mov dh, [ext_support]
    ; Jump to stage 2
    jmp STAGE2_SEGMENT:0000

disk_error:
    ; Set video mode to text mode
    mov ax, 0x0003
    int 0x10
    
    ; Set blue background (1) and white text (F)
    mov ax, 0x0B00
    mov bx, 0x001F    ; Blue background (1), intensity (8), white text (F)
    int 0x10
    
    ; Clear screen with blue background
    mov ax, 0x0600
    mov bh, 0x1F      ; Blue background (1), white text (F)
    mov cx, 0x0000    ; Upper left corner
    mov dx, 0x184F    ; Lower right corner
    int 0x10
    
    ; Move cursor to top row
    mov ah, 0x02
    mov bh, 0
    mov dx, 0x0000    ; Row 0, Column 0
    int 0x10
    
    ; Print error messages
    mov si, stage1_err_code
    call print_string
    mov si, stage1_disk_error
    call print_string
    mov si, newline
    call print_string
    mov si, stage1_halt
    call print_string
    
    cli
    hlt
    jmp $

print_string:
    .next_char:
        lodsb
        cmp al, 0
        je .done
        mov ah, 0x0E
        int 0x10
        jmp .next_char
    .done:
        ret

; Data
boot_drive db 0
ext_support db 0
stage1_err_code   db "ERR 0x1", 0x0D, 0x0A, 0
stage1_disk_error db "Couldn't read disk.", 0x0D, 0x0A, 0
stage1_halt       db "System halted.", 0x0D, 0x0A, 0
newline          db 0x0D, 0x0A, 0

; Disk Address Packet for extended reads
dap:
    db 0x10        ; Size of packet
    db 0           ; Reserved
    dw STAGE2_SECTORS  ; Number of sectors
    dw 0           ; Transfer buffer offset
    dw STAGE2_SEGMENT ; Transfer buffer segment
    dq 1           ; Starting LBA (sector 2)

times 510-($-$$) db 0
dw 0xAA55
