; stage2.asm
[BITS 16]
[ORG 0]

KERNEL_SEGMENT equ 0x1000
NORMAL_ATTR equ 0x07    ; Light gray (7) on black (0)
SELECT_ATTR equ 0x70    ; Black (0) on light gray (7)
KERNEL_SECTORS equ 64

start:
    ; Set up segment registers for stage 2
    mov ax, cs
    mov ds, ax
    mov es, ax
    
    ; Store boot drive and ext support flag passed from stage 1
    mov [boot_drive], dl
    mov [ext_support], dh

    ; Reset disk system first
    xor ah, ah
    mov dl, [boot_drive]
    int 0x13
    jc disk_error

    ; Clear the screen and setup display
    mov ax, 0x0003      ; Switch to text mode 3
    int 0x10
    
    mov ax, 0x0700
    mov bh, 0x07
    mov cx, 0
    mov dx, 0x184F
    int 0x10

    ; Disable cursor
    mov ah, 0x01    ; Set cursor shape function
    mov ch, 0x20    ; Bit 5 set = disable cursor
    int 0x10

    ; Move cursor to top left
    mov ah, 0x02
    mov bh, 0
    mov dx, 0
    int 0x10

    ; Display boot message
    mov si, boot_msg
    call print_string

    ; Display boot drive info
    mov si, booting_from_msg
    call print_string

    ; Determine and print drive type
    mov al, [boot_drive]
    cmp byte [ext_support], 1
    je .print_usb
    cmp al, 0x80
    jae .print_hd
    cmp al, 0x00
    je .print_fda
    mov si, drive_type_b
    jmp .print_type

.print_usb:
    mov si, drive_type_usb
    jmp .print_type

.print_hd:
    mov si, drive_type_hd
    jmp .print_type

.print_fda:
    mov si, drive_type_a

.print_type:
    call print_string
    mov si, newline
    call print_string
    mov si, newline    ; Extra newline for spacing
    call print_string

show_menu:
    ; Initialize selected option
    mov byte [selected_option], 0

    ; Store cursor position for menu start
    mov ah, 0x03
    xor bh, bh
    int 0x10
    mov [menu_row], dh

    ; Print initial menu
    call draw_menu

.menu_loop:
    ; Wait for keypress
    xor ah, ah
    int 0x16

    ; Check for arrow keys
    cmp ah, 0x48    ; Up arrow
    je .up_pressed
    cmp ah, 0x50    ; Down arrow
    je .down_pressed
    cmp al, 0x0D    ; Enter
    je .enter_pressed
    jmp .menu_loop

.up_pressed:
    mov al, [selected_option]
    test al, al
    jz .menu_loop    ; If already at top
    dec al
    mov [selected_option], al
    call draw_menu
    jmp .menu_loop

.down_pressed:
    mov al, [selected_option]
    cmp al, 1        ; Number of options - 1
    jae .menu_loop   ; If already at bottom
    inc al
    mov [selected_option], al
    call draw_menu
    jmp .menu_loop

.enter_pressed:
    mov al, [selected_option]
    test al, al
    jz normal_boot
    jmp restart_system

; Draw menu with current selection highlighted
draw_menu:
    pusha
    ; Save cursor position
    mov ah, 0x03
    xor bh, bh
    int 0x10
    push dx

    ; Move to menu start position
    mov ah, 0x02
    mov dh, [menu_row]
    mov dl, 0
    int 0x10

    ; Print menu header
    mov si, menu_msg
    call print_string
    mov si, newline
    call print_string

    ; Print "Normal Mode" option
    mov bl, NORMAL_ATTR
    cmp byte [selected_option], 0
    jne .print_normal
    mov bl, SELECT_ATTR
.print_normal:
    mov si, normal_mode_msg
    call print_string_attr
    mov si, newline
    call print_string

    ; Print "Restart System" option
    mov bl, NORMAL_ATTR
    cmp byte [selected_option], 1
    jne .print_restart
    mov bl, SELECT_ATTR
.print_restart:
    mov si, restart_mode_msg
    call print_string_attr
    mov si, newline
    call print_string

    ; Restore cursor position
    pop dx
    mov ah, 0x02
    int 0x10
    popa
    ret

; Print string with attribute in BL
print_string_attr:
    push ax
.next_char:
    lodsb
    test al, al
    jz .done
    mov ah, 0x09
    push bx
    push cx
    mov cx, 1
    int 0x10
    pop cx
    pop bx
    ; Move cursor forward
    mov ah, 0x03
    push bx
    xor bh, bh
    int 0x10
    pop bx
    inc dl
    mov ah, 0x02
    int 0x10
    jmp .next_char
.done:
    mov ah, 0x02
    int 0x10
    pop ax
    ret

disk_error:
    mov ax, 0x0003      ; Switch to text mode 3
    int 0x10
    
    mov ax, 0x0B00
    mov bx, 0x001F
    int 0x10
    
    mov ax, 0x0600
    mov bh, 0x1F
    mov cx, 0x0000
    mov dx, 0x184F
    int 0x10
    
    mov ah, 0x02
    mov bh, 0
    mov dx, 0x0000
    int 0x10
    
    mov si, bsod_err_code
    call print_string
    mov si, bsod_disk_error
    call print_string
    mov si, newline
    call print_string
    mov si, bsod_halt
    call print_string
    cli
    hlt
    jmp $

print_string:
    .next_char:
        lodsb
        cmp al, 0
        je .done
        mov ah, 0x0E
        int 0x10
        jmp .next_char
    .done:
        ret

normal_boot:
    mov byte [boot_mode], 0
    jmp load_kernel

restart_system:
    jmp 0xFFFF:0x0000

load_kernel:
    pusha               ; Save all registers
    
    ; Check if using extended disk access
    cmp byte [ext_support], 1
    je .ext_load

    ; Legacy kernel load with retry mechanism
    mov cx, 3          ; Retry counter
.retry_legacy:
    push cx            ; Save retry counter
    
    mov ax, KERNEL_SEGMENT
    mov es, ax
    xor bx, bx        ; ES:BX = KERNEL_SEGMENT:0
    mov ah, 0x02      ; Read sectors
    mov al, KERNEL_SECTORS
    mov ch, 0         ; Cylinder 0
    mov cl, 4         ; Start from sector 4
    mov dh, 0         ; Head 0
    mov dl, [boot_drive]
    int 0x13
    jnc .load_complete
    
    ; On error, reset disk and retry
    xor ah, ah
    int 0x13
    pop cx
    loop .retry_legacy
    jmp disk_error    ; If all retries failed

.ext_load:
    mov cx, 3         ; Retry counter
.retry_ext:
    push cx           ; Save retry counter
    
    ; Set up disk address packet
    mov word [dap.count], KERNEL_SECTORS
    mov word [dap.offset], 0
    mov word [dap.segment], KERNEL_SEGMENT
    mov dword [dap.lba_low], 3    ; Start from sector 4 (LBA 3)
    mov dword [dap.lba_high], 0
    
    mov si, dap
    mov ah, 0x42
    mov dl, [boot_drive]
    int 0x13
    jnc .load_complete
    
    ; On error, reset disk and retry
    xor ah, ah
    int 0x13
    pop cx
    loop .retry_ext
    jmp disk_error    ; If all retries failed

.load_complete:
    popa                    ; Restore all registers
    
    ; Set up segment registers for kernel
    mov ax, KERNEL_SEGMENT
    mov ds, ax
    mov es, ax
    
    ; Pass boot info to kernel
    mov dl, [boot_drive]
    mov dh, [ext_support]
    
    ; Jump to kernel
    jmp KERNEL_SEGMENT:0x0000

; Data
boot_mode       db 0
boot_drive      db 0
selected_option db 0
menu_row        db 0
ext_support     db 0
boot_msg        db "ExploreOS Bootloader Version 0.45 [0.45.0028]", 13, 10, 0
booting_from_msg db "Booting from: ", 0
drive_type_a    db "Floppy Drive A:", 0
drive_type_b    db "Floppy Drive B:", 0
drive_type_hd   db "Hard Drive", 0
drive_type_usb  db "USB Drive", 0
drive_type_unknown db "Unknown Drive", 0
newline         db 13, 10, 0
menu_msg        db "Boot Menu:", 0
normal_mode_msg  db "1. Normal Mode", 0
restart_mode_msg db "2. Restart System", 0
bsod_err_code   db "ERR 0x1", 13, 10, 0
bsod_disk_error db "Disk read error!", 13, 10, 0
bsod_halt       db "System halted.", 13, 10, 0

dap:
    .size       db 0x10
    .reserved   db 0
    .count      dw 0
    .offset     dw 0
    .segment    dw 0
    .lba_low    dd 0
    .lba_high   dd 0
